
#include <stdio.h>
#include <stdlib.h>
#include "stl.h"
#include "ch3d.h"

extern void	Error(char *msg);

typedef struct {
		float	x,y,z;
	} Vertex;

static double	x1,y1,z1,x2,y2,z2;

static void	Store_and_Check(Vertex *v,float x,float y,float z)
{
	if ((v->x=x)<x1) x1=x; else if (x>x2) x2=x;
	if ((v->y=y)<y1) y1=y; else if (y>y2) y2=y;
	if ((v->z=z)<z1) z1=z; else if (z>z2) z2=z;
}

static int	Compare(Vertex *v1,Vertex *v2)
{
	return (v1->x<v2->x)?-1:(v1->x>v2->x)?1:
	       (v1->y<v2->y)?-1:(v1->y>v2->y)?1:
	       (v1->z<v2->z)?-1:(v1->z>v2->z)?1:0;
}

static void	PutN(FILE *file,char *mem,int num)
{
	(void)fwrite(mem,(size_t)4,(size_t)num,file);
}

static void	PutR(FILE *file,char *mem,int num)
{
	char	c[4];

	for (; --num>=0; mem+=4) {
	    c[0]=mem[3]; c[1]=mem[2]; c[2]=mem[1]; c[3]=mem[0]; PutN(file,c,1);
	}
}

#define SVB	4096	/* size of vertex buffer */
#define EPS	1e-5
#define WID	((long double)CH3D_MAX-(long double)CH3D_MIN)

int	main(int argc,char **argv)
{
	FILE		*file;
	STL		stl;
	float		f[12];
	CH3D_U		q,p[3],o=0,n=0;
	Vertex		*v=NULL;
	double		dx,dy,dz,x0,y0,z0,x,y,z;
	CH3D_T		*ch3d;
	unsigned	u;
	short		s=1;
	char		*c=(char *)&s;
	void		(*Put)()=((short)*c==s)?PutN:PutR;

	if (argc!=3) Error("usage : stl_ch orgSTL newSTL");

	if (argv[1][0]=='-')
	    file=stdin;
	else
	    if ((file=fopen(argv[1],"rb"))==NULL) Error("file not found.");

	if (!SetUpSTL(&stl,file)) Error("header read error.");

	while (ScanSTL(&stl,f+0,f+1,f+2,f+3,f+4,f+5,f+6,f+7,f+8)>0) {
	    if (o%(SVB*3)==0 &&
		(v=(Vertex *)realloc(v,sizeof(Vertex)*(o+SVB*3)))==NULL)
		Error("memory allocation error.");

	    if (o==0) {
		x1=x2=f[0]; y1=y2=f[1]; z1=z2=f[2];
	    }
	    Store_and_Check(v+o++,f[0],f[1],f[2]);
	    Store_and_Check(v+o++,f[3],f[4],f[5]);
	    Store_and_Check(v+o++,f[6],f[7],f[8]);
	}
	if ((dx=x2-x1)<EPS || (dy=y2-y1)<EPS || (dz=z2-z1)<EPS)
	    Error("too narrow data range.");

	if (file!=stdin) (void)fclose(file);

	qsort(v,(size_t)o,sizeof(Vertex),(int (*)())Compare);

	for (q=1; q<o; q++)
	    if (v[n].x!=v[q].x || v[n].y!=v[q].y || v[n].z!=v[q].z) v[++n]=v[q];

	if ((v=(Vertex *)realloc(v,sizeof(Vertex)*(++n)))==NULL)
	    Error("memory re-allocation error.");

	ch3d=InitCH3D(n);

	x0=(x1+x2)/2.0; y0=(y1+y2)/2.0; z0=(z1+z2)/2.0;
	for (q=0; q<n; q++) {
	    CH3D_X(ch3d)[q]=(CH3D_I)(WID*((v[q].x-x0)/dx));
	    CH3D_Y(ch3d)[q]=(CH3D_I)(WID*((v[q].y-y0)/dy));
	    CH3D_Z(ch3d)[q]=(CH3D_I)(WID*((v[q].z-z0)/dz));
        }
	CH3D(ch3d,n,0);

	if (argv[2][0]=='-')
	    file=stdout;
	else
	    if ((file=fopen(argv[2],"wb"))==NULL) Error("file not stored.");

	for (q=0; q<80; q++) (void)fputc(0,file);

	u=(unsigned)ch3d->Nf1; (*Put)(file,&u,1);

	for (q=0; q<ch3d->Nf1; q++) {
	    CH3D_f2p(ch3d,q,p);
	    CalcNormal((double)(f[ 3]=v[p[0]].x),
		       (double)(f[ 4]=v[p[0]].y),
		       (double)(f[ 5]=v[p[0]].z),
		       (double)(f[ 6]=v[p[1]].x),
		       (double)(f[ 7]=v[p[1]].y),
		       (double)(f[ 8]=v[p[1]].z),
		       (double)(f[ 9]=v[p[2]].x),
		       (double)(f[10]=v[p[2]].y),
		       (double)(f[11]=v[p[2]].z),&x,&y,&z);
	    f[0]=x; f[1]=y; f[2]=z; (*Put)(file,f,12);

	    (void)fputc(0,file); (void)fputc(0,file);
	}
	if (file!=stdout) (void)fclose(file);

	TermCH3D(ch3d); free(v); return 0;
}
