
#include <stdio.h>
#include <stdlib.h>
#include "stl.h"

#define INT_WID	((1<<28)-1)	/* must be less than 1<<28 */
#define XYZ_BLK	(2048*3)	/* must be multiplier of 3 */

extern void	Error(),chull3_();

typedef struct {
		float	x,y,z;
	} XYZ;

static int	Compare(xyz1,xyz2)
XYZ		*xyz1,*xyz2;
{
	return (xyz1->x<xyz2->x)?-1:(xyz1->x>xyz2->x)?1:
	       (xyz1->y<xyz2->y)?-1:(xyz1->y>xyz2->y)?1:
	       (xyz1->z<xyz2->z)?-1:(xyz1->z>xyz2->z)?1:0;
}

static int	FV3(X,Y,Z,fe,sv,ev,rf,sce,ece,f,v1,v2,v3)
int		*X,*Y,*Z,*fe,*sv,*ev,*rf,*sce,*ece,f,*v1,*v2,*v3;
{
	int		e=fe[f]-1;
	long long	x1,y1,z1,x2,y2,z2;

	if (f==rf[e]-1) {
	    *v1=ev[e]-1; *v2=sv[e]-1; e=sce[e]-1;
	}
	else {
	    *v1=sv[e]-1; *v2=ev[e]-1; e=ece[e]-1;
	}
	*v3=(f==rf[e]-1)?sv[e]-1:ev[e]-1;

	return (x1=X[*v1]-X[*v3])*(y2=Y[*v2]-Y[*v3])!=
	       (y1=Y[*v1]-Y[*v3])*(x2=X[*v2]-X[*v3]) ||
	       (z1=Z[*v1]-Z[*v3])*x2!=x1*(z2=Z[*v2]-Z[*v3]) || y1*z2!=z1*y2;
}

static void	PutN(file,ptr)
FILE		*file;
char		*ptr;
{
	(void)fwrite(ptr,(size_t)4,(size_t)1,file);
}

static void	PutR(file,ptr)
FILE		*file;
char		*ptr;
{
	char	c[4];

	c[0]=ptr[3]; c[1]=ptr[2]; c[2]=ptr[1]; c[3]=ptr[0]; PutN(file,c);
}

static void	Put0(file,cnt)
FILE		*file;
int		cnt;
{
	while (--cnt>=0) (void)fputc(0,file);
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	FILE	*file;
	STL	stl;
	XYZ	*xyz;
	int	V,v,i,N1,N2,N3,F,v1,v2,v3,
		*x,*y,*z,*name,para[10],
		*ve,*fe,*sv,*ev,*rf,*lf,*sce,*scce,*ece,*ecce,
		*wv1,*wv2,*wv3,*wv4,*wv5,*wv6,*we1,*we2,*wf1,*wf2;
	float	f[12];
	double	a,b,nx,ny,nz;

	short	s=1;
	char	*c=(char *)&s;
	void	(*Put)()=((short)*c==s)?PutN:PutR;

	if (argc!=3) Error("usage : stl_ch orgSTL newSTL");

	if (argv[1][0]=='-')
	    file=stdin;
	else
	    if ((file=fopen(argv[1],"rb"))==NULL) Error("file not found.");

	if (!SetUpSTL(&stl,file)) Error("header read error.");

	xyz=NULL; V=0;
	while (ScanSTL(&stl,f+0,f+1,f+2,f+3,f+4,f+5,f+6,f+7,f+8)>0) {
	    if (V%XYZ_BLK==0 && 
	       (xyz=(XYZ *)realloc(xyz,sizeof(XYZ)*(V+XYZ_BLK)))==NULL)
		Error("no allocatable memory for original points.");

	    xyz[V].x=f[0]; xyz[V].y=f[1]; xyz[V].z=f[2]; ++V;
	    xyz[V].x=f[3]; xyz[V].y=f[4]; xyz[V].z=f[5]; ++V;
	    xyz[V].x=f[6]; xyz[V].y=f[7]; xyz[V].z=f[8]; ++V;
	}
	if (file!=stdin) (void)fclose(file);

	qsort(xyz,(size_t)V,sizeof(XYZ),(int (*)())Compare);

	v=0;
	for (i=1; i<V; i++)
	    if (xyz[v].x!=xyz[i].x ||
		xyz[v].y!=xyz[i].y ||
		xyz[v].z!=xyz[i].z) if (++v!=i) xyz[v]=xyz[i];

	V=v+1; N1=V+1; N2=N1*3; N3=N1*4;

#define M1(x)	(x=(int *)malloc(sizeof(int)*N1))==NULL
#define M2(fe)	(fe=(int *)malloc(sizeof(int)*N2))==NULL
#define M3(sv)	(sv=(int *)malloc(sizeof(int)*N3))==NULL

	if ((xyz=(XYZ *)realloc(xyz,sizeof(XYZ)*V))==NULL ||
	    M1(x) || M1(y) || M1(z) || M1(name) ||
	    M1(ve) || M2(fe) || M3(sv) || M3(ev) || M3(rf) || M3(lf) ||
	    M3(sce) || M3(scce) || M3(ece) || M3(ecce) ||
	    M1(wv1) || M1(wv2) || M1(wv3) || M1(wv4) || M1(wv5) || M1(wv6) ||
	    M3(we1) || M3(we2) || M2(wf1) || M2(wf2))
	    Error("no allocatable memory for convex hull.");

	for (i=0; i<V; i++) {
	    if (i==0) a=b=xyz[0].x; else
	    if (xyz[i].x<a) a=xyz[i].x; else if (xyz[i].x>b) b=xyz[i].x;
	    if (xyz[i].y<a) a=xyz[i].y; else if (xyz[i].y>b) b=xyz[i].y;
	    if (xyz[i].z<a) a=xyz[i].z; else if (xyz[i].z>b) b=xyz[i].z;
	}
	b=(double)(2*INT_WID)/(b-a); a=0.5-a*b;
	for (i=0; i<V; i++) {
	    x[i]=(int)(xyz[i].x*b+a)-INT_WID;
	    y[i]=(int)(xyz[i].y*b+a)-INT_WID;
	    z[i]=(int)(xyz[i].z*b+a)-INT_WID; name[i]=i;
	}
	para[0]=N1; chull3_(&V,x,y,z,name,para,
			    ve,fe,sv,ev,rf,lf,sce,scce,ece,ecce,
			    wv1,wv2,wv3,wv4,wv5,wv6,we1,we2,wf1,wf2);

	if (argv[2][0]=='-')
	    file=stdout;
	else
	    if ((file=fopen(argv[2],"wb"))==NULL) Error("file not stored.");

	Put0(file,80);

	F=0;
	for (i=0; i<N2; i++)
	    F+=(wf1[i] && FV3(x,y,z,fe,sv,ev,rf,sce,ece,i,&v1,&v2,&v3));

	(*Put)(file,&F);

	for (i=0; i<N2; i++)
	    if (wf1[i] && FV3(x,y,z,fe,sv,ev,rf,sce,ece,i,&v1,&v2,&v3)) {
		CalcNormal((double)(f[ 3]=xyz[v1].x),
			   (double)(f[ 4]=xyz[v1].y),
			   (double)(f[ 5]=xyz[v1].z),
			   (double)(f[ 6]=xyz[v2].x),
			   (double)(f[ 7]=xyz[v2].y),
			   (double)(f[ 8]=xyz[v2].z),
			   (double)(f[ 9]=xyz[v3].x),
			   (double)(f[10]=xyz[v3].y),
			   (double)(f[11]=xyz[v3].z),&nx,&ny,&nz); f[0]=nx;
								   f[1]=ny;
								   f[2]=nz;

		for (v=0; v<12; v++) (*Put)(file,f+v);

		Put0(file,2);
	    }

	if (file!=stdout) (void)fclose(file);

	free(wf2); free(wf1); free(we2); free(we1);
	free(wv6); free(wv5); free(wv4); free(wv3); free(wv2); free(wv1);
	free(ecce); free(ece); free(scce); free(sce);
	free(lf); free(rf); free(ev); free(sv); free(fe); free(ve);
	free(name); free(z); free(y); free(x);
	free(xyz);

	return 0;
}
