
#include <stdio.h>
#include <stdlib.h>
#include <string.h>	/* using strcmp() */
#include "ch3d.h"

extern void	Error(char *msg);

static void	GetN(FILE *file,char *mem)
{
	if (fread(mem,(size_t)4,(size_t)1,file)!=1) Error("data read error.");
}

static void	GetR(FILE *file,char *mem)
{
	char	c[4];

	GetN(file,c); mem[3]=c[0]; mem[2]=c[1]; mem[1]=c[2]; mem[0]=c[3];
}

static void	PutN(FILE *file,char *mem)
{
	(void)fwrite(mem,(size_t)4,(size_t)1,file);
}

static void	PutR(FILE *file,char *mem)
{
	char	c[4];

	c[0]=mem[3]; c[1]=mem[2]; c[2]=mem[1]; c[3]=mem[0]; PutN(file,c);
}

#define EPS	1e-5
#define WID	((long double)CH3D_MAX-(long double)CH3D_MIN)

int	main(int argc,char **argv)
{
	FILE		*file;
	unsigned long	v,f;
	short		s=1;
	char		e_h[12],
			*c=(char *)&s;
	void		(*Get)()=(s==(short)*c)?GetN:GetR,
			(*Put)()=(s==(short)*c)?PutN:PutR;
	float		*X,*Y,*Z;
	CH3D_U		i,p,P[3];
	double		x1,x2,y1,y2,z1,z2,dx,dy,dz,x0,y0,z0;
	CH3D_T		*ch3d;
	unsigned	u;

	if (argc!=3) Error("usage : zcp_ch orgZCP newZCP");

	if (argv[1][0]=='-')
	    file=stdin;
	else
	    if ((file=fopen(argv[1],"rb"))==NULL) Error("file not found.");

	if (fscanf(file,"ply\n"
			"format binary_little_endian 1.0\n"
			"element vertex %lu\n"
			"property float x\n"
			"property float y\n"
			"property float z\n"
			"element face %lu\n"
			"property uchar red\n"
			"property uchar green\n"
			"property uchar blue\n"
			"property list uchar int vertex_indices\n",&v,&f)!=2 ||
	    fgets(e_h,12,file)==NULL || strcmp(e_h,"end_header\n"))
	    Error("bad PLY/ZCP header.");

	if ((X=(float *)malloc(sizeof(float)*v))==NULL ||
	    (Y=(float *)malloc(sizeof(float)*v))==NULL ||
	    (Z=(float *)malloc(sizeof(float)*v))==NULL)
	    Error("memory allocation error.");

	for (i=0; i<v; i++) {
	    (*Get)(file,X+i); (*Get)(file,Y+i); (*Get)(file,Z+i);

	    if (i==0) {
		x1=x2=X[0]; y1=y2=Y[0]; z1=z2=Z[0];
	    }
	    else {
		if (x1>X[i]) x1=X[i]; else if (x2<X[i]) x2=X[i];
		if (y1>Y[i]) y1=Y[i]; else if (y2<Y[i]) y2=Y[i];
		if (z1>Z[i]) z1=Z[i]; else if (z2<Z[i]) z2=Z[i];
	    }
	}
	if ((dx=x2-x1)<EPS || (dy=y2-y1)<EPS || (dz=z2-z1)<EPS)
	    Error("too narrow data range.");

	if (file!=stdin) (void)fclose(file);

	ch3d=InitCH3D((CH3D_U)v);

	x0=(x1+x2)/2.0; y0=(y1+y2)/2.0; z0=(z1+z2)/2.0;
	for (i=0; i<v; i++) {
	    CH3D_X(ch3d)[i]=(CH3D_I)(WID*((X[i]-x0)/dx));
	    CH3D_Y(ch3d)[i]=(CH3D_I)(WID*((Y[i]-y0)/dy));
	    CH3D_Z(ch3d)[i]=(CH3D_I)(WID*((Z[i]-z0)/dz));
	}
	CH3D(ch3d,(CH3D_U)v,0);

	if (argv[2][0]=='-')
	    file=stdout;
	else
	    if ((file=fopen(argv[2],"wb"))==NULL) Error("file not stored.");

	(void)fprintf(file,"ply\n"
			   "format binary_little_endian 1.0\n"
			   "element vertex %lu\n"
			   "property float x\n"
			   "property float y\n"
			   "property float z\n"
			   "element face %lu\n"
			   "property uchar red\n"
			   "property uchar green\n"
			   "property uchar blue\n"
			   "property list uchar int vertex_indices\n"
			   "end_header\n",(unsigned long)ch3d->Nv1,
					  (unsigned long)ch3d->Nf1);

	for (i=0; i<ch3d->Nv1; i++) {
	    p=CH3D_V2P(ch3d,i);
	    (*Put)(file,X+p); (*Put)(file,Y+p); (*Put)(file,Z+p);
	}
	for (i=0; i<ch3d->Nf1; i++) {
	    u=0x3ffffff; (*Put)(file,&u);

	    CH3D_f2p(ch3d,i,P);
	    for (p=0; p<3; p++) {
		u=(unsigned)CH3D_P2V(ch3d,P[p]); (*Put)(file,&u);
	    }
	}
	if (file!=stdout) (void)fclose(file);

	TermCH3D(ch3d); free(Z); free(Y); free(X); return 0;
}
